/*
** element.c - Element lookup
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "pjc.h"

CHANGER *
changer_open(const char *path)
{
    CHANGER *scp;
    unsigned char buf[255];
    size_t buflen;
    int rc;
    

    scp = malloc(sizeof(*scp));
    memset(scp, 0, sizeof(*scp));

    scp->scsi = scsi_open(path);
    if (!scp->scsi)
	goto Fail;

    /* Get SCSI Element Address Assignment Page */
    rc = scsi_mode_sense_6(scp->scsi, 0, 0, 0x1D, buf, sizeof(buf), &buflen);
    if (rc != 0)
	goto Fail;
    
    if (buflen != 24)
	goto Fail;
	

    scp->config.transport.start = ((buf[6]<<8)|buf[7]);
    scp->config.transport.count = ((buf[8]<<8)|buf[9]);
    
    scp->config.storage.start = ((buf[10]<<8)|buf[11]);
    scp->config.storage.count = ((buf[12]<<8)|buf[13]);
    
    scp->config.ioport.start = ((buf[14]<<8)|buf[15]);
    scp->config.ioport.count = ((buf[16]<<8)|buf[17]);
    
    scp->config.drive.start = ((buf[18]<<8)|buf[19]);
    scp->config.drive.count = ((buf[20]<<8)|buf[21]);

    return scp;
    
  Fail:
    if (scp->scsi)
	scsi_close(scp->scsi);
    
    free(scp);
    return NULL;
}

int
changer_close(CHANGER *scp)
{
    int rc;
    
    rc = scsi_close(scp->scsi);
    free(scp);

    return rc;
}




int
changer_voltag_equal(const unsigned char *s1,
		     const char *s2)
{
    int s1len, s2len;

    s1len = 32;
    while (s1len > 0 && isspace(s1[s1len-1]))
	--s1len;

    s2len = strlen(s2);
    while (s2len > 0 && isspace((unsigned) s2[s2len-1]))
	--s2len;

    if (s1len != s2len)
	return 0;

    return strncmp((const char *) s1, s2, s1len) == 0;
}


int
changer_element_lookup(CHANGER *scp,
		       char *obj)
{
    static unsigned char *buf = NULL;
    static size_t buflen = 0;
    static int n_elems = 0;
    char *cp;
    int n, i, j, elemtypes, rc, p, id, addr, elem_t, full;
    

    if (sscanf(obj, "#%u", &id) == 1 ||
	sscanf(obj, "%u", &id) == 1)
	return id;

    cp = strchr(obj, ':');
    if (!cp)
    {
	elem_t = 0; /* Lookup by label ID */
    }
    else
    {
	int len;

	
	++cp;
	len = cp-obj;
	
	if (strncmp(obj, "transport:", len) == 0)
	    elem_t = 1;
	else if (strncmp(obj, "storage:", len) == 0)
	    elem_t = 2;
	else if (strncmp(obj, "ioport:", len) == 0)
	    elem_t = 3;
	else if (strncmp(obj, "drive:", len) == 0)
	    elem_t = 4;
	else
	    return -1;
	
	if (sscanf(cp, "%u", &id) != 1)
	    return -1;
    }


    if (buf == NULL)
    {
	n_elems = scp->config.transport.count+
	    scp->config.storage.count+
	    scp->config.ioport.count+
	    scp->config.drive.count;
	
	buflen = 1024+n_elems*256;
	buf = malloc(buflen);
	if (!buf)
	    return -1;
	
	rc = scsi_read_element_status(scp->scsi, 0, 0, n_elems, buf, buflen, &buflen);
	if (rc != 0)
	    goto Fail;
    }
	

    p = 0;
    elemtypes = (buf[p+2]<<8) + buf[p+3];
    
    p += 8;
    for (i = 0; i < elemtypes;)
    {
	size_t elength, dlength;
	int etype, pvoltag, avoltag;
	
	etype = buf[p];
	pvoltag = (buf[p+1]&0x80);
	avoltag = (buf[p+1]&0x40);
	
	elength = (buf[p+2]<<8) + buf[p+3];
	dlength = (buf[p+5]<<16) + (buf[p+6]<<8) + buf[p+7];
	
	p += 8;

	n = element_start;
	
	for (j = 0; j < dlength; j += elength, p += elength, n++)
	{
	    int xp;
	    
	    ++i;

	    addr = (buf[p+0]<<8) + buf[p+1];
	    full = (buf[p+2] & 0x01);
	    
	    if (elem_t > 0 && elem_t == etype && n == id)
		goto Done;

	    if (!elem_t)
		switch (etype)
		{
		  case 1: /* Medium Transport Element */
		    xp = p+12;
		    if (pvoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
			    
			xp += 36;
		    }
		    if (avoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
		    }
		    break;
		    
		  case 2: /* Storage Element */
		    xp = p+12;
		    if (pvoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
			xp += 36;
		    }
		    if (avoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
		    }
		    break;

		  case 3: /* Import/Export Element */
		    xp = p+12;
		    if (pvoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
			xp += 36;
		    }
		    if (avoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
		    }
		    break;
		    
		  case 4: /* Data Transfer Element */
		    xp = p+12;
		    if (pvoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
			xp += 36;
		    }
		    if (avoltag)
		    {
			if (changer_voltag_equal(buf+xp, obj))
			    goto Done;
		    }
		    break;
		}
	}
    }

#if 0
    free(buf);
#endif
    return 0;
    
  Fail:
#if 0
    free(buf);
#endif
    return -1;
    
  Done:
#if 0
    free(buf);
#endif
    return addr;

}

