/*
** commands.c - command line commands
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>

#include "pjc.h"


static void
spin(void)
{
    static int state = 0;
    static const char dial[] = "|/-\\";

    putchar(dial[state%(sizeof(dial)-1)]);
    putchar('\b');

    ++state;
}


int
cmd_test_unit_ready(int argc,
		    char *argv[])
{
    return scsi_test_unit_ready(changer->scsi);
}


int
cmd_rezero_unit(int argc,
		char *argv[])

{
    return scsi_rezero_unit(changer->scsi);
}

int
cmd_init_element_status(int argc,
			char *argv[])
{
    return scsi_initialize_element_status(changer->scsi);
}

int
cmd_request_sense(int argc,
		  char *argv[])
{
    int rc;
    SCSI_SENSE sbuf;
    
    
    rc = scsi_request_sense(changer->scsi, &sbuf);
    if (rc == 0)
    {
	if (sbuf.len < 18)
	{
	    fprintf(stderr,
		    "%s: scsi_request_sense: invalid SENSE data length: %u\n",
		    argv[0], (unsigned int) sbuf.len);
	    return CMD_ERROR;
	}

	if (verbose)
	    print_sense_verbose(stdout, &sbuf);
	else
	    print_sense(stdout, &sbuf);
	
	return CMD_OK;
    }

    return rc;
}

int
cmd_inquiry(int argc,
	    char *argv[])
{
    unsigned char *ib;
    size_t iplen = 0;
    int ver;
    

    ver = SCSI_INQUIRY_VERSION(&changer->scsi->inquiry);
    
    ib = changer->scsi->inquiry.buf;
    iplen = changer->scsi->inquiry.len;
    
    if (verbose)
    {
	printf("  Peripheral Qualifier:      ");
	switch ((ib[0]>>5)&0x07)
	{
	  case 0:
	    puts("Device (probably) available");
	    break;
	    
	  case 1:
	    puts("Device currently not available");
	    break;
	    
	  case 2:
	    puts("Reserved");
	    break;
	    
	  case 3:
	    puts("Device not supported");
	    break;
	    
	  default:
	    printf("Vendor Specific [0x%x]\n", ((ib[0]>>5)&0x07));
	    break;
	}
	
	printf("  Device type:               %s\n", scsi_devicetype2string(ib[0]&0x1F));
	printf("  Removable:                 %s\n", (ib[1]&0x80) ? "Yes" : "No");
	
	if (changer->scsi->version != ver)
	    printf("  SCSI Version Override:     %d (%s)\n",
		   changer->scsi->version, scsi_ansiversion2string(changer->scsi->version));
	
	if (ver > 2)
	    printf("  SCSI Version:              %d\n", ver);
	else
	{
	    printf("  ISO Version:               %d\n", (ib[2]>>6)&0x03);
	    printf("  ECMA Version:              %d\n", (ib[2]>>3)&0x07);
	    printf("  ANSI Version:              %s\n", scsi_ansiversion2string(ver));
	}
	printf("  Manufacturer:              %.8s\n", ib+8);
	printf("  Product:                   %.16s\n", ib+16);
	printf("  Revision:                  %.4s\n", ib+32);
	printf("  Response Data Format:      %u\n", (ib[3]&0x0F));
	
	
	if (changer->scsi->version > 2)
	{
	    /* SCSI-3 fields */
	    
	    printf("  Normal ACA:                %s\n", (ib[3]&0x20) ? "Supported" : "Not supported");
	    printf("  Hierarchical Addressing:   %s\n", (ib[3]&0x10) ? "Supported" : "Not supported");
	    
	    printf("  SCC Supported:             %s\n", (ib[5]&0x80) ? "Yes" : "No");
	    printf("  AC Coordinator:            %s\n", (ib[5]&0x40) ? "Yes" : "No");
	    printf("  Target Port Group Support: %u\n", ((ib[5]>>4)&0x03));
	    printf("  Third-Party Copy:          %s\n", (ib[5]&0x08) ? "Supported" : "Not supported");
	    printf("  Protect Information:       %s\n", (ib[5]&0x01) ? "Supported" : "Not supported");
	    
	    printf("  BQueue Flag:               %s\n", (ib[6]&0x80) ? "On" : "Off");
	    printf("  Enclosure Services:        %s\n", (ib[6]&0x40) ? "Available" : "Not available");
	    printf("  Multiport Device:          %s\n", (ib[6]&0x10) ? "Yes" : "No");
	    printf("  Attached Media Changer:    %s\n", (ib[6]&0x08) ? "Yes" : "No");
	    printf("  Addr16 Flag:               %s\n", (ib[6]&0x01) ? "On" : "Off");
	    
	    printf("  WBus16 Flag:               %s\n", (ib[7]&0x20) ? "On" : "Off");
	    printf("  Sync Flag:                 %s\n", (ib[7]&0x10) ? "On" : "Off");
	    printf("  Linked Flag:               %s\n", (ib[7]&0x08) ? "On" : "Off");
	    printf("  CmdQue Flag:               %s\n", (ib[7]&0x02) ? "On" : "Off");
	}
	else if (changer->scsi->version == 2)
	{
	    printf("  RelAdr Flag:               %s\n", (ib[7]&0x80) ? "On" : "Off");
	    printf("  WBus32 Flag:               %s\n", (ib[7]&0x40) ? "On" : "Off");
	    printf("  WBus16 Flag:               %s\n", (ib[7]&0x20) ? "On" : "Off");
	    printf("  Sync Flag:                 %s\n", (ib[7]&0x10) ? "On" : "Off");
	    printf("  Linked Flag:               %s\n", (ib[7]&0x08) ? "On" : "Off");
	    printf("  CmdQue Flag:               %s\n", (ib[7]&0x02) ? "On" : "Off");
	    printf("  SftRe Flag:                %s\n", (ib[7]&0x01) ? "On" : "Off");
	}
    }
    else
    {
	printf("device-type:   %u (%s)\n", ib[0]&0x1F,
	       scsi_devicetype2string(ib[0]&0x1F));
	
	printf("removable:     %s\n", (ib[1]&0x80) ? "yes" : "no");
	printf("manufacturer:  %.8s\n", ib+8);
	printf("product:       %.16s\n", ib+16);
	printf("revision:      %.4s\n", ib+32);
    }
    
    return CMD_OK;
}

int
cmd_reserve(int argc,
	    char *argv[])
{
    int i, n;
    unsigned int xid, xidflag, rvid, rvidflag;
    unsigned char *evbuf = NULL;
    size_t evsize = 0;


    xidflag = rvidflag = 0;
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	switch (argv[i][1])
	{
	  case 'S':
	    if (sscanf(argv[i]+2, "%u", &xid) != 1)
	    {
		fprintf(stderr, "%s: %s: invalid SCSI id number: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	    xidflag = 1;
	    break;

	  case 'R':
	    if (sscanf(argv[i]+2, "0x%x", &rvid) != 1 &&
		sscanf(argv[i]+2, "%u", &rvid) != 1)
	    {
		fprintf(stderr, "%s: %s: invalid reservation id number: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	    rvidflag = 1;
	    break;

	  default:
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }

    if (i < argc)
    {
	evsize = 4+2*(argc-i);
	if (debug)
	    fprintf(stderr, "cmd_reserve: evsize=%u (n=%u)\n",
		    (unsigned int) evsize, argc-i);
	
	evbuf = malloc(evsize);
	if (!evbuf)
	{
	    fprintf(stderr, "%s: %s: malloc(%u): %s\n",
		    argv0, argv[0], (unsigned int) evsize, strerror(errno));
	    return CMD_ERROR;
	}

	evbuf[0] = evbuf[1] = 0;
	evbuf[2] = (((argc-i)>>8)&0xFF);
	evbuf[3] = ((argc-i)&0xFF);

	n = 0;
	while (i < argc)
	{
	    unsigned int addr = changer_element_lookup(changer, argv[i]);
	    if (addr <= 0)
	    {
		fprintf(stderr, "%s: %s: invalid element: %s\n",
			argv0, argv[0], argv[i]);
		return CMD_ERROR;
	    }
	    
	    evbuf[4+n*2+0] = ((addr>>8)&0xFF);
	    evbuf[4+n*2+1] = (addr&0xFF);
	    
	    ++i;
	    ++n;
	}
    }
		
    return scsi_reserve(changer->scsi, xidflag ? &xid : NULL, rvidflag ? &rvid : NULL, evbuf, evsize);
}


int
cmd_release(int argc,
	    char *argv[])
{
    int i;
    unsigned int xid, xidflag, rvid, rvidflag;


    xidflag = rvidflag = 0;
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	switch (argv[i][1])
	{
	  case 'S':
	    if (sscanf(argv[i]+2, "%u", &xid) != 1)
	    {
		fprintf(stderr, "%s: %s: invalid SCSI id number: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	    xidflag = 1;
	    break;

	  case 'R':
	    if (sscanf(argv[i]+2, "0x%x", &rvid) != 1 &&
		sscanf(argv[i]+2, "%u", &rvid) != 1)
	    {
		fprintf(stderr, "%s: %s: invalid reservation id number: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	    rvidflag = 1;
	    break;

	  default:
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }

    if (i < argc)
    {
	fprintf(stderr, "%s: %s: extra arguments (ignored):",
		argv0, argv[0]);
	while (i < argc)
	    fprintf(stderr, " %s", argv[i++]);
    }
    
    return scsi_release(changer->scsi, xidflag ? &xid : NULL, rvidflag ? &rvid : NULL);
}


int
cmd_move_medium(int argc,
		char *argv[])
{
    unsigned int transport, source, dest, invert;
    int i;
    

    transport = source = dest = invert = 0;

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	if (strcmp(argv[i], "-I") == 0)
	    invert = 1;

	else if (strncmp(argv[i], "-T", 2) == 0)
	{
	    transport = changer_element_lookup(changer, argv[i]+2);
	    if (transport <= 0)
	    {
		fprintf(stderr, "%s: %s: invalid transport: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	}
	
	else
	{
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n", argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }
    
    if (i >= argc)
    {
 	fprintf(stderr, "%s: %s: missing source address\n",
 		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    source = changer_element_lookup(changer, argv[i]);
    if (source <= 0)
    {
	fprintf(stderr, "%s: %s: invalid source address: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }
    
    ++i;
    
    if (i >= argc)
    {
	fprintf(stderr, "%s: %s: missing destination address\n",
		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    dest = changer_element_lookup(changer, argv[i]);
    if (dest <= 0)
    {
	fprintf(stderr, "%s: %s: invalid destination address: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }
    
    ++i;
    
    if (i < argc)
    {
	fprintf(stderr, "%s: %s: extra arguments (ignored):",
		argv0, argv[0]);
	while (i < argc)
	    fprintf(stderr, " %s", argv[i++]);
    }
    
    return scsi_move_medium(changer->scsi, transport, source, dest, invert);
}


int
cmd_mode_sense_6(int argc,
		 char *argv[])
{
    unsigned char buf[255];
    size_t buflen;
    int dbd = 0, pc = 0, page_code, rc, i;


    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	if (strncmp(argv[i], "-P", 2) == 0)
	{
	    if (sscanf(argv[i]+2, "%u", &pc) != 1)
	    {
		if (strcmp(argv[i]+2, "current") == 0)
		    pc = 0;
		else if (strcmp(argv[i]+2, "changeable") == 0)
		    pc = 1;
		else if (strcmp(argv[i]+2, "default") == 0)
		    pc = 2;
		else if (strcmp(argv[i]+2, "saved") == 0)
		    pc = 3;
		else
		{
		    fprintf(stderr, "%s: %s: invalid page control value: %s\n",
			    argv0, argv[0], argv[i]+2);
		    return CMD_ERROR;
		}
	    }
	}
	
	else if (strcmp(argv[i], "-D") == 0)
	    dbd = 1;
	
	else
	{
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }
    
    if (i >= argc)
    {
	fprintf(stderr, "%s: %s: missing page code\n",
		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    if (sscanf(argv[i], "0x%x", &page_code) != 1 &&
	sscanf(argv[i], "%u", &page_code) != 1)
    {
	fprintf(stderr, "%s: %s: invalid page code: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }
    
    rc = scsi_mode_sense_6(changer->scsi, dbd, pc, page_code, buf, sizeof(buf), &buflen);
    if (rc == 0)
    {
	if (buflen < 4)
	{
	    fprintf(stderr, "%s: %s: invalid size of sense data: %u\n",
		    argv0, argv[0], (unsigned int) buflen);
	    return CMD_ERROR;
	}
	
	print_mode_page_parameter_header(stdout, buf);
	putc('\n', stdout);
	
	switch (page_code)
	{
	  case SCSI_MODE_PAGE_ELEMENT_ADDRESS_ASSIGNMENT:
	    print_mode_page_element_address_assignment(stdout, buf+4, buflen-4);
	}

	return CMD_OK;
    }

    return rc;
}

int
cmd_list_element_status(int argc,
			char *argv[])
{
    unsigned int start = 0;
    unsigned int elems = 0;
    size_t bufsize, buflen;
    unsigned int ecode = 0;
    unsigned char *buf;
    int i, rc;
	

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	if (strncmp(argv[i], "-N", 2) == 0)
	{
	    if (sscanf(argv[i]+2, "%u", &elems) != 1)
	    {
		fprintf(stderr, "%s: %s: invalid number of elements: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	}

	else if (strncmp(argv[i], "-E", 2) == 0)
	{
	    if (sscanf(argv[i]+2, "%u", &ecode) != 1)
	    {
		if (strcmp(argv[i]+2, "transport") == 0)
		    ecode = 1;
		else if (strcmp(argv[i]+2, "storage") == 0)
		    ecode = 2;
		else if (strcmp(argv[i]+2, "ioport") == 0)
		    ecode = 3;
		else if (strcmp(argv[i]+2, "drive") == 0)
		    ecode = 4;
		else
		{
		    fprintf(stderr, "%s: %s: invalid element code: %s\n",
			    argv0, argv[0], argv[i]+2);
		    return CMD_ERROR;
		}
	    }
	    
	}

	else
	{
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }

    if (!elems)
    {
	elems = changer->config.transport.count +
	    changer->config.storage.count +
	    changer->config.ioport.count +
	    changer->config.drive.count;
    }

    
    bufsize = elems * 256 + 1024; /* XXX: Fixme! What amount should be allocated per element? */
    buflen = 0;
    buf = malloc(bufsize);	
    if (!buf)
    {
	fprintf(stderr, "%s: %s: malloc(%u): %s\n",
		argv0, argv[0], (unsigned int) bufsize, strerror(errno));
	return CMD_ERROR;
    }

    rc = scsi_read_element_status(changer->scsi, ecode, start, elems, buf, bufsize, &buflen);
    if (rc == 0)
    {
	if (i >= argc)
	{
	    if (verbose)
	    {
		print_element_status_verbose(stdout, buf, buflen, 0);
	    }
	    else
		print_element_status(stdout, buf, buflen, 0);
	}
	else
	{
	    while (i < argc)
	    {
		start = changer_element_lookup(changer, argv[i]);
		if (start <= 0)
		{
		    fprintf(stderr, "%s: %s: invalid element address: %s\n",
			    argv0, argv[0], argv[i]);
		    return CMD_ERROR;
		}
		
		if (verbose)
		    print_element_status_verbose(stdout, buf, buflen, start);
		else
		    print_element_status(stdout, buf, buflen, start);
		
		++i;
	    }
	}
    }

    free(buf);
    return rc;
}


int
cmd_show_element_assignment(int argc,
			    char *argv[])
{
    if (verbose)
	printf("Medium Changer Element Address Assignments:\n");
	
    printf("  Medium Transport Element Start: %u\n", changer->config.transport.start);
    printf("  Medium Transport Element Count: %u\n", changer->config.transport.count);
    printf("  Storage Element Start:          %u\n", changer->config.storage.start);
    printf("  Storage Element Count:          %u\n", changer->config.storage.count);
    printf("  Import/Export Element Start:    %u\n", changer->config.ioport.start);
    printf("  Import/Export Element Count:    %u\n", changer->config.ioport.count);
    printf("  Data Transport Element Start:   %u\n", changer->config.drive.start);
    printf("  Data Transport Element Count:   %u\n", changer->config.drive.count);
    
    return CMD_OK;
}

int
cmd_show_transport_geometry(int argc,
			    char *argv[])
{
    unsigned char buf[255];
    size_t buflen;
    int rc, i;


    rc = scsi_mode_sense_6(changer->scsi, 0, 0, 0x1E, buf, sizeof(buf), &buflen);
    if (rc != 0)
	return rc;

    if (buflen < 6)
    {
	fprintf(stderr, "%s: %s: invalid sense buffer size returned: %u\n",
		argv0, argv[0], (unsigned int) buflen);
	return CMD_ERROR;
    }

    
    printf("  Parameters Savable: %s\n", (buf[4]&0x80) ? "Yes" : "No");
    printf("  Transports:\n");
    for (i = 0; i < buf[4+1]; i += 2)
	printf("    %4u\t%s\n", buf[6+i+1], (buf[6+i]&1) ? "Rotatable" : "Fixed");

    return CMD_OK;
}

int
cmd_show_device_capabilities(int argc,
			     char *argv[])
{
    unsigned char buf[255];
    size_t buflen;
    int rc;


    rc = scsi_mode_sense_6(changer->scsi, 0, 0, 0x1F, buf, sizeof(buf), &buflen);
    if (rc != 0)
	return rc;

    if (buflen != 24)
    {
	fprintf(stderr, "%s: %s: invalid sense buffer size returned: %u\n",
		argv0, argv[0], (unsigned int) buflen);
	return CMD_ERROR;
    }

    
    printf("  Parameters Savable: %s\n", (buf[4+0]&0x80) ? "Yes" : "No");
    putchar('\n');

    printf("  Capabilities:\n");
    printf("    Storage  : Drive:                    %s\n", (buf[4+2]&0x08) ? "Yes" : "No");
    printf("    Storage  : IOport:                   %s\n", (buf[4+2]&0x04) ? "Yes" : "No");
    printf("    Storage  : Storage:                  %s\n", (buf[4+2]&0x02) ? "Yes" : "No");
    printf("    Storage  : Transport:                %s\n", (buf[4+2]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Move     : Transport -> Drive:       %s\n", (buf[4+4]&0x08) ? "Yes" : "No");
    printf("    Move     : Transport -> IOPort:      %s\n", (buf[4+4]&0x04) ? "Yes" : "No");
    printf("    Move     : Transport -> Storage:     %s\n", (buf[4+4]&0x02) ? "Yes" : "No");
    printf("    Move     : Transport -> Transport:   %s\n", (buf[4+4]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Move     : Storage -> Drive:         %s\n", (buf[4+5]&0x08) ? "Yes" : "No");
    printf("    Move     : Storage -> IOPort:        %s\n", (buf[4+5]&0x04) ? "Yes" : "No");
    printf("    Move     : Storage -> Storage:       %s\n", (buf[4+5]&0x02) ? "Yes" : "No");
    printf("    Move     : Storage -> Transport:     %s\n", (buf[4+5]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Move     : IOPort -> Drive:          %s\n", (buf[4+6]&0x08) ? "Yes" : "No");
    printf("    Move     : IOPort -> IOPort:         %s\n", (buf[4+6]&0x04) ? "Yes" : "No");
    printf("    Move     : IOPort -> Storage:        %s\n", (buf[4+6]&0x02) ? "Yes" : "No");
    printf("    Move     : IOPort -> Transport:      %s\n", (buf[4+6]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Move     : Drive -> Drive:           %s\n", (buf[4+7]&0x08) ? "Yes" : "No");
    printf("    Move     : Drive -> IOPort:          %s\n", (buf[4+7]&0x04) ? "Yes" : "No");
    printf("    Move     : Drive -> Storage:         %s\n", (buf[4+7]&0x02) ? "Yes" : "No");
    printf("    Move     : Drive -> Transport:       %s\n", (buf[4+7]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Exchange : Transport <> Drive:       %s\n", (buf[4+12]&0x08) ? "Yes" : "No");
    printf("    Exchange : Transport <> IOPort:      %s\n", (buf[4+12]&0x04) ? "Yes" : "No");
    printf("    Exchange : Transport <> Storage:     %s\n", (buf[4+12]&0x02) ? "Yes" : "No");
    printf("    Exchange : Transport <> Transport:   %s\n", (buf[4+12]&0x01) ? "Yes" : "No");
    putchar('\n');

    printf("    Exchange : Storage <> Drive:         %s\n", (buf[4+13]&0x08) ? "Yes" : "No");
    printf("    Exchange : Storage <> IOPort:        %s\n", (buf[4+13]&0x04) ? "Yes" : "No");
    printf("    Exchange : Storage <> Storage:       %s\n", (buf[4+13]&0x02) ? "Yes" : "No");
    printf("    Exchange : Storage <> Transport:     %s\n", (buf[4+13]&0x01) ? "Yes" : "No");
    putchar('\n');
    
    printf("    Exchange : IOPort <> Drive:          %s\n", (buf[4+14]&0x08) ? "Yes" : "No");
    printf("    Exchange : IOPort <> IOPort:         %s\n", (buf[4+14]&0x04) ? "Yes" : "No");
    printf("    Exchange : IOPort <> Storage:        %s\n", (buf[4+14]&0x02) ? "Yes" : "No");
    printf("    Exchange : IOPort <> Transport:      %s\n", (buf[4+14]&0x01) ? "Yes" : "No");
    putchar('\n');
    
    printf("    Exchange : Drive <> Drive:           %s\n", (buf[4+15]&0x08) ? "Yes" : "No");
    printf("    Exchange : Drive <> IOPort:          %s\n", (buf[4+15]&0x04) ? "Yes" : "No");
    printf("    Exchange : Drive <> Storage:         %s\n", (buf[4+15]&0x02) ? "Yes" : "No");
    printf("    Exchange : Drive <> Transport:       %s\n", (buf[4+15]&0x01) ? "Yes" : "No");
    
    return CMD_OK;
}

int
cmd_changer_element_lookup(int argc,
			   char *argv[])
{
    int i, rc = -EINVAL;

    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	;
    
    if (i >= argc)
    {
	fprintf(stderr, "%s: %s: missing object(s) to lookup\n",
		argv0, argv[0]);
	return CMD_ERROR;
    }

    while (i < argc)
    {
	rc = changer_element_lookup(changer, argv[i]);
	if (rc > 0)
	{
	    if (verbose)
		printf("Element '%s' has address %u\n", argv[i], rc);
	    else
		printf("%-32s %4u\n", argv[i], rc);
	    rc = 0;
	}
	
	else if (rc == 0)
	{
	    if (verbose)
		printf("Element '%s' - not found\n", argv[i]);
	    else
		printf("%-32s %4s\n", argv[i], "?");
	}
	
	else
	    break;
	
	++i;
    }

    return rc;
}
    


int
cmd_position_to_element(int argc,
			char *argv[])
{
    unsigned int transport, dest, invert;
    int i;
    

    transport = dest = invert = 0;

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	if (strcmp(argv[i], "-I") == 0)
	    invert = 1;
	
	else
	{
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n", argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }
    
    if (i >= argc)
    {
	fprintf(stderr, "%s: %s: missing destination address\n",
		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    dest = changer_element_lookup(changer, argv[i]);
    if (dest <= 0)
    {
	fprintf(stderr, "%s: %s: invalid destination address: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }
    
    ++i;
    
    if (i < argc)
    {
	if (sscanf(argv[i], "%u", &transport) != 1)
	{
	    fprintf(stderr, "%s: %s: invalid transport: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }

    ++i;
    
    if (i < argc)
    {
	fprintf(stderr, "%s: %s: extra arguments (ignored):",
		argv0, argv[0]);
	while (i < argc)
	    fprintf(stderr, " %s", argv[i++]);
    }
    
    return scsi_position_to_element(changer->scsi, transport, dest, invert);
}



int
cmd_exchange_medium(int argc,
		    char *argv[])
{
    unsigned int transport, source, dest1, dest2, invert;
    int i;
    

    transport = source = dest1 = dest2 = invert = 0;

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
    {
	if (strcmp(argv[i], "-I1") == 0)
	    invert |= 0x01;
	
	else if (strcmp(argv[i], "-I2") == 0)
	    invert |= 0x02;
	
	else if (strncmp(argv[i], "-T", 2) == 0)
	{
	    transport = changer_element_lookup(changer, argv[i]+2);
	    if (transport <= 0)
	    {
		fprintf(stderr, "%s: %s: invalid transport: %s\n",
			argv0, argv[0], argv[i]+2);
		return CMD_ERROR;
	    }
	}
	
	else
	{
	    fprintf(stderr, "%s: %s: invalid sub-option: %s\n", argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
    }
    
    if (i >= argc)
    {
 	fprintf(stderr, "%s: %s: missing source address\n",
 		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    source = changer_element_lookup(changer, argv[i]);
    if (source <= 0)
    {
	fprintf(stderr, "%s: %s: invalid source address: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }

    dest2 = source;
    
    ++i;
    
    if (i >= argc)
    {
	fprintf(stderr, "%s: %s: missing destination-1 address\n",
		argv0, argv[0]);
	return CMD_ERROR;
    }
    
    dest1 = changer_element_lookup(changer, argv[i]);
    if (dest1 <= 0)
    {
	fprintf(stderr, "%s: %s: invalid destination-1 address: %s\n",
		argv0, argv[0], argv[i]);
	return CMD_ERROR;
    }

    ++i;

    if (i < argc)
    {    
	dest2 = changer_element_lookup(changer, argv[i]);
	if (dest2 <= 0)
	{
	    fprintf(stderr, "%s: %s: invalid destination-2 address: %s\n",
		    argv0, argv[0], argv[i]);
	    return CMD_ERROR;
	}
	
	++i;
    }
    
    if (i < argc)
    {
	fprintf(stderr, "%s: %s: extra arguments (ignored):",
		argv0, argv[0]);
	while (i < argc)
	    fprintf(stderr, " %s", argv[i++]);
    }
    
    return scsi_exchange_medium(changer->scsi, transport, source, dest1, dest2, invert);
}


int
cmd_prevent_removal(int argc,
		    char *argv[])
{
    return scsi_prevent_allow_removal(changer->scsi, 1);
}


int
cmd_allow_removal(int argc,
		  char *argv[])
{
    return scsi_prevent_allow_removal(changer->scsi, 0);
}


static struct command
{
    const char *name;
    int (*handler)(int argc, char *argv[]);
    const char *usage;
} cmdv[] = {
    
    {
	"test-unit-ready",
	cmd_test_unit_ready,
	""
    },
    
    {
	"rezero-unit",
	cmd_rezero_unit,
	""
    },
    
    {
	"initialize-element-status",
	cmd_init_element_status,
	""
    },
    
    {
	"request-sense",
	cmd_request_sense,
	""
    },
    
    {
	"inquiry",
	cmd_inquiry,
	""
    },
    
    {
	"reserve",
	cmd_reserve,
	"[-S<scsi-id>] [-R<reservation-id>] [<element1> [... <elementN>]]"
    },
    
    {
	"release",
	cmd_release,
	"[-S<scsi-id>] [-R<reservation-id>]"
    },
    
    {
	"move-medium",
	cmd_move_medium,
	"[-I] [-T<transport>] <src> <dst>"
    },
    
    {
	"mode-sense-6",
	cmd_mode_sense_6,
	"[-P{current|changeable|default|saved}] [-D] <pagecode>"
    },
    
    {
	"list-element-status",
	cmd_list_element_status,
	"[-N<nelems>] [-E{transport|storage|ioport|drive}] <element1> [... <elementN>]"
    },
    
    {
	"show-element-address-assignment",
	cmd_show_element_assignment,
	""
    },
    
    {
	"show-transport-geometry",
	cmd_show_transport_geometry,
	""
    },
    
    {
	"show-device-capabilities",
	cmd_show_device_capabilities,
	""
    },
    
    {
	"lookup-element",
	cmd_changer_element_lookup,
	"<element>"
    },
    
    {
	"position-to-element",
	cmd_position_to_element,
	"[-I] <element>"
    },
    
    {
	"exchange-medium",
	cmd_exchange_medium,
	"[-I{1|2}] [-T<transport>] <src> <dst-1> [<dst-2>]"
    },

    {
	"prevent-removal",
	cmd_prevent_removal,
	""
    },
    
    {
	"allow-removal",
	cmd_allow_removal,
	""
    },
    
    {
	"help",
	cmd_help,
	""
    },
    
    { NULL, NULL }
};


static int
cmd_locate(const char *name,
	   struct command **cp)
{
    int i;
    int nmatch = 0;
    

    for (i = 0; cmdv[i].name; i++)
    {
	if (strncmp(name, cmdv[i].name, strlen(name)) == 0)
	{
	    *cp = &cmdv[i];
	    ++nmatch;
	}
    }

    return nmatch;
}


int
cmd_help(int argc,
	 char *argv[])
{
    int i, j;
    

    if (argc <= 1)
    {
	printf("Available commands:\n");
	for (i = 0; cmdv[i].name; i++)
	    printf("  %s\n", cmdv[i].name);
    }
    else
    {
	for (i = 1; i < argc; i++)
	{
	    int nmatch = 0;
	    
	    printf("Command usage:\n");
	    for (j = 0; cmdv[j].name; j++)
	    {
		if (strncmp(cmdv[j].name, argv[i], strlen(argv[i])) == 0)
		{
		    ++nmatch;
		    if (cmdv[j].name)
			printf("  %s %s\n", cmdv[j].name, cmdv[j].usage);
		}
	    }

	    if (!nmatch)
	    {
		fprintf(stderr, "  %s: unknown command\n", argv[i]);
		return 1;
	    }
	}
    }

    return 0;
}


int
cmd_execute(int argc,
	    char *argv[])
{
    int nm;
    struct command *cp = NULL;
    

    nm = cmd_locate(argv[0], &cp);
    if (nm < 1)
    {
	fprintf(stderr, "%s: invalid command: %s\n", argv0, argv[0]);
	return CMD_ERROR;
    }
    else if (nm > 1)
    {
	fprintf(stderr, "%s: command not unique: %s\n", argv0, argv[0]);
	return CMD_ERROR;
    }
	
    return cp->handler(argc, argv);
}
