/*
** pjc.c - Peter's Jukebox Controller
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>

#include "pjc.h"


char *argv0 = "pjc";

int timeout = 0;
int verbose = 0;
int debug = 0;
int pktdump = 0;
int waitretry = 0;
int element_start = 1;

CHANGER *changer = NULL;

char *device_name = NULL;



int
get_time(const char *buf,
	 int *sec)
{
    char c = '\0';
    int rc;

    
    rc = sscanf(buf, "%d%c", sec, &c);
    if (rc < 1)
	return rc;

    if (rc == 2)
	switch (c)
	{
	  case 's':
	    break;
	  case 'm':
	    *sec *= 60;
	    break;
	  case 'h':
	    *sec *= 60*60;
	    break;
	  case 'd':
	    *sec *= 60*60*24;
	    break;
	  case 'w':
	    *sec *= 60*60*24*7;
	    break;
	  case 'M':
	    *sec *= 60*60*24*31;
	    break;
	  case 'Q':
	    *sec *= 60*60*24*31*3;
	    break;
	  case 'Y':
	    *sec *= 60*60*24*31*12;
	    break;
	  default:
	    return -1;
	}

    return 1;
}


int
get_size(const char *buf,
	 int *size)
{
    char c = '\0';
    int rc;

    
    rc = sscanf(buf, "%d%c", size, &c);
    if (rc < 1)
	return rc;

    if (rc == 2)
	switch (c)
	{
	  case 'k':
	    *size *= 1024;
	    break;
	  case 'M':
	    *size *= 1024*1024;
	    break;
	  case 'G':
	    *size *= 1024*1024*1024;
	    break;
	  default:
	    return -1;
	}

    return 1;
}


int
main(int argc,
     char *argv[])
{
    int rc, i, j, nexti;
    char *devmap, *cp;
    int s_timeout = 0;
    int s_version = 0;
    int s_attached = 0;

    argv0 = argv[0];
    
    if ((device_name = getenv("CHANGER")) == NULL)
	device_name = "/dev/changer";
    
    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
    {
	nexti = 0;
	for (j = 1; !nexti && argv[i][j]; ++j)
	    switch (argv[i][j])
	    {
	      case 'H':
		exit(cmd_help(argc-i, argv+i));
		
	      case 'h':
		printf("Usage: %s [global options] <command> [<command options>] [<command args>]\n",
		       argv0);
		puts("Global options:\n");
		puts("  --              End of global options");
		puts("  -h              Display this information");
		puts("  -H [cmd]        Display all commands / usage of command");
		puts("  -w [<cnt>]      Wait for ready before executing command");
		puts("  -f <dev>        Changer device");
		puts("  -D <obj>=<dev>  Define device mapping");
		puts("  -s <off>        Device numbering start");
		puts("  -t <sec>        SCSI Command timeout [seconds]");
		puts("  -T <sec>        Global timeout [seconds]");
		puts("  -A [<lev>]      Force attached changer [level]");
		puts("  -V <lev>        Force SCSI command version [level]");
		puts("  -d [<lev>]      Increase/[Set] debug information [level]");
		puts("  -v [<lev>]      Increase/[Set] verbose level [level]");
		puts("  -x [<len>]      Display SCSI packet data/parts of [bytes]");
		exit(0);
		
	      case '-':
		break;

	      case 'D':
		if (argv[i][j+1])
		{
		    devmap = strdup(argv[i]+j+1);
		    nexti = 1;
		}
		else if (i+1 < argc)
		{
		    devmap = strdup(argv[++i]);
		    nexti = 1;
		}
		else
		{
		    fprintf(stderr, "%s: -%c: missing device-map argument\n", argv[0], argv[i][j]);
		    exit(1);
		}

		cp = strchr(devmap, '=');
		if (!cp)
		{
		    fprintf(stderr, "%s: -%c: invalid device-map argument: %s\n", argv[0], argv[i][j], devmap);
		    exit(1);
		}
		
		*cp++ = '\0';
		if (devmap_insert(devmap, cp) < 0)
		{
		    *--cp = '=';
		    fprintf(stderr, "%s: -%c: invalid device-map argument: %s\n", argv[0], argv[i][j], devmap);
		    exit(1);
		}
		break;

		
	      case 'f':
		if (argv[i][j+1])
		{
		    device_name = strdup(argv[i]+j+1);
		    nexti = 1;
		}
		else if (i+1 < argc)
		{
		    device_name = strdup(argv[++i]);
		    nexti = 1;
		}
		else
		{
		    fprintf(stderr, "%s: missing device argument for '-f'\n", argv[0]);
		    exit(1);
		}
		break;

	      case 'T':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (get_time(argv[i]+j+1, &timeout) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-T': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && get_time(argv[i+1], &timeout) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else 
		    timeout = -1;
		break;

	      case 'A':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &s_attached) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-A': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &s_attached) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else 
		    s_attached = -1;
		break;

	      case 'V':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &s_version) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-V': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &s_version) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else 
		{
		    fprintf(stderr, "%s: missing required argument for '-V'\n", argv[0]);
		    exit(1);
		}
		break;

	      case 't':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (get_time(argv[i]+j+1, &s_timeout) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-t': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && get_time(argv[i+1], &s_timeout) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    s_timeout = -1;
		break;

	      case 'd':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &debug) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-d': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &debug) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    ++debug;
		break;

	      case 's':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &element_start) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-s': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &element_start) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    element_start = 0;
		break;

	      case 'w':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &waitretry) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-w': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &waitretry) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    waitretry = -1; /* No limit */
		break;
		
	      case 'v':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (sscanf(argv[i]+j+1, "%u", &verbose) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-v': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && sscanf(argv[i+1], "%u", &verbose) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    ++verbose;
		break;

	      case 'x':
		if (argv[i][j+1] && isdigit((unsigned) argv[i][j+1]))
		{
		    if (get_size(argv[i]+j+1, &pktdump) != 1)
		    {
			fprintf(stderr, "%s: invalid argument for '-x': %s\n", argv[0], argv[i]+j+1);
			exit(1);
		    }
		    nexti = 1;
		}
		else if (!argv[i][j+1] && i+1 < argc && get_size(argv[i+1], &pktdump) == 1)
		{
		    ++i;
		    nexti = 1;
		}
		else
		    pktdump = -1;
		break;

	      default:
		fprintf(stderr, "%s: unknown switch '%c' in argument '%s'\n",
			argv[0], argv[i][j], argv[i]);
		exit(1);
	    }
    }

    if (verbose)
	printf("[PJC %s - Copyright (c) Peter Eriksson - Built %s %s]\n\n",
	       version, __DATE__, __TIME__);

    if (timeout > 0)
	alarm(timeout);
    
    changer = changer_open(device_name);
    if (!changer)
    {
	fprintf(stderr, "%s: changer_open: %s: %s\n",
		argv[0], device_name, strerror(errno));
	exit(1);
    }

    if (s_timeout)
	scsi_timeout(changer->scsi, s_timeout);
    
    if (s_version)
	changer->scsi->version = s_version;
    if (s_attached)
	changer->scsi->attached = s_attached;
    

    rc = cmd_execute(argc-i, argv+i);
    if (rc == CMD_OK)
    {
	if (verbose > 1)
	    printf("Command '%s' executed successfully.\n", argv[i]);

	exit(0);
    }

    if (rc < CMD_OK)
    {
	fprintf(stderr, "%s: %s: failed with code %u: %s\n",
		argv[0], argv[i], -rc, strerror(-rc));
    }
    
    else if (rc < CMD_ERROR)
    {
	const char *ss = scsi_status2string(rc);

	
	if (rc == SCSI_STATUS_CHECK_CONDITION ||
	    rc == SCSI_STATUS_COMMAND_TERMINATED)
	{
	    if (verbose)
	    {
		fprintf(stderr, "%s: %s: failed: SCSI Status 0x%02x (Check Condition) : Sense Data:\n",
			argv[0], argv[i], rc);
		
		print_sense(stderr, &changer->scsi->sense);
	    }
	    else
	    {
		const char *sks = scsi_sense_key2string(SCSI_SENSE_KEY(&changer->scsi->sense));
		const char *scs = scsi_sense_code2string(SCSI_SENSE_ASC(&changer->scsi->sense),
							 SCSI_SENSE_ASCQ(&changer->scsi->sense));

		fprintf(stderr, "%s: %s: failed: ", argv[0], argv[i]);
		if (sks)
		    fprintf(stderr, "%s", sks);
		else
		    fprintf(stderr, "sense key 0x%02x", SCSI_SENSE_KEY(&changer->scsi->sense));
		if (scs)
		    fprintf(stderr, " (%s)\n", scs);
		else
		    fprintf(stderr, " (0x%02x / 0x%02x)",
			    SCSI_SENSE_ASC(&changer->scsi->sense),
			    SCSI_SENSE_ASCQ(&changer->scsi->sense));
			
	    }
	   
	}
	else
	    fprintf(stderr, "%s: %s: failed: SCSI Status 0x%02x (%s)\n",
		    argv[0], argv[i],
		    rc,
		    ss ? ss : "Unknown");
    }
    
    exit(1);
}
