/*
** print.c - Print various data structures
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <ctype.h>

#include "pjc.h"


void
print_buffer(FILE *fp,
	     const char *header,
	     void *buffer,
	     size_t buflen)
{
    unsigned char *bp = (unsigned char *) buffer;
    unsigned pos = 0, i;

    
    fprintf(fp, "%s (size=%u):\n", header, (unsigned int) buflen);
    if (!buffer)
	return;
    
    while (pos < buflen)
    {
	fprintf(fp, "  %04x :", pos);
	
	for (i = 0; i < 16 && pos+i < buflen; i++)
	{
	    if (i == 8)
		putc(' ', fp);
	    fprintf(fp, " %02x", bp[pos+i]);
	}
	
	while (i < 16)
	{
	    if (i == 8)
		putc(' ', fp);
	    fputs("   ", fp);
	    ++i;
	}
	
	fputs(" : ", fp);
	
	for (i = 0; i < 16 && pos+i < buflen; i++)
	{
	    if (i == 8)
		putc(' ', fp);
	    if (isprint(bp[pos+i]))
		putc(bp[pos+i], fp);
	    else
		putc('?', fp);
	}
	
	putc('\n', fp);
	pos += i;
    }
}


void
print_sense_verbose(FILE *fp,
		    SCSI_SENSE *ssp)
{
    const char *ecs = scsi_sense_error2string(SCSI_SENSE_ERROR(ssp));
    const char *sks = scsi_sense_key2string(SCSI_SENSE_KEY(ssp));
    const char *scs = scsi_sense_code2string(SCSI_SENSE_ASC(ssp),SCSI_SENSE_ASCQ(ssp));
    unsigned int len;
    unsigned char *rqs = ssp->buf;
    int valid = (rqs[0] & 0x80);
    
    fprintf(fp, "Error Code:                    0x%02x (%s)\n",
	    SCSI_SENSE_ERROR(ssp),
	    ecs ? ecs : "Unknown");

    fprintf(fp, "Segment Number:                0x%02x\n", rqs[1]);
    
    fprintf(fp, "File Mark:                     %s\n", (rqs[2]&0x80) ? "Yes" : "No");
    fprintf(fp, "End of Media:                  %s\n", (rqs[2]&0x40) ? "Yes" : "No");
    fprintf(fp, "ILI:                           %s\n", (rqs[2]&0x20) ? "Yes" : "No");
    
    fprintf(fp, "Sense Key:                     0x%02x (%s)\n",
	    SCSI_SENSE_KEY(ssp),
	    sks ? sks : "Unknown");

    fprintf(fp, "Information Field Valid:       %s\n", valid ? "Yes" : "No");
	    
    fprintf(fp, "Information:                   0x%02x 0x%02x 0x%02x 0x%02x\n",
	    rqs[3], rqs[4], rqs[5], rqs[6]);

    len = rqs[7];

    fprintf(fp, "Command-specific Information:  0x%02x 0x%02x 0x%02x 0x%02x\n",
	    rqs[3], rqs[4], rqs[5], rqs[6]);

    fprintf(fp, "Sense Code / Extended:         0x%02x / 0x%02x (%s)\n",
	    SCSI_SENSE_ASC(ssp),
	    SCSI_SENSE_ASCQ(ssp),
	    scs ? scs : "Unknown");

    fprintf(fp, "FRU:                           0x%02x\n",
	    SCSI_SENSE_FRU(ssp));

    if (rqs[15]&0x80)
    {
	fprintf(stderr, "Sense Key Specific Information:\n");
	switch (rqs[2])
	{
	  case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
	    fprintf(fp, "   C/D:                        %s\n", ((rqs[15]>>6)&1) ? "Yes" : "No");
	    if ((rqs[15]>>3)&1)
		fprintf(fp, "   Bit Position:               %u\n", (rqs[15]&0x07));
	    fprintf(fp, "   Error Position:             %u\n", ((rqs[16]<<8)|rqs[17]));
	    break;

	  default:
	    fprintf(fp, "   Byte 15:                    0x%02x\n", rqs[15]);
	    fprintf(fp, "   Byte 16:                    0x%02x\n", rqs[16]);
	    fprintf(fp, "   Byte 17:                    0x%02x\n", rqs[17]);
	}

	if (len > 0)
	    print_buffer(fp, "Additional Sense Bytes", rqs+18, len); 
    }
}

void
print_sense(FILE *fp,
	    SCSI_SENSE *ssp)
{
    unsigned char *rqs = ssp->buf;
    const char *ecs = scsi_sense_error2string((rqs[0]&0x7F));
    const char *sks = scsi_sense_key2string(rqs[2]&0x0F);
    const char *scs = scsi_sense_code2string(rqs[12],rqs[13]);
    

    fprintf(fp, "error-code:  0x%02x (%s)\n", (rqs[0]&0x7F),
	    ecs ? ecs : "Unknown");

    fprintf(fp, "sense-key:   0x%02x (%s)\n", (rqs[2]&0x0F),
	    sks ? sks : "Unknown");

    fprintf(fp, "sense-code:  0x%02x / 0x%02x (%s)\n",
	    rqs[12], rqs[13],
	    scs ? scs : "Unknown");

    fprintf(fp, "fru:         0x%02x\n", rqs[14]);
}


void
print_element_status_verbose(FILE *fp,
			     unsigned char *buf,
			     size_t buflen,
			     unsigned int element)
{
    int p = 0;
    int i, j, elemtypes, n, addr, seq;
    
    
    elemtypes = (buf[p+2]<<8) + buf[p+3];

    if (!element)
    {
	fprintf(fp, "First element:       %u\n", (buf[p+0]<<8) + buf[p+1]);
	fprintf(fp, "Number of elements:  %u\n", elemtypes);
    }
    
    p += 8;
    for (i = 0; i < elemtypes;)
    {
	size_t elength, dlength;
	int etype, pvoltag, avoltag;
	
	etype = buf[p];
	pvoltag = (buf[p+1]&0x80);
	avoltag = (buf[p+1]&0x40);

	if (!element)
	{
	    putc('\n', fp);
	    fprintf(fp, "  Element type code: %u (%s)\n", etype, scsi_element_type2string(etype));
	}
	
	elength = (buf[p+2]<<8) + buf[p+3];
	dlength = (buf[p+5]<<16) + (buf[p+6]<<8) + buf[p+7];
	
	p += 8;
	
	n = element_start;

	for (j = 0; j < dlength; j += elength, p += elength, n++)
	{
	    int xp;
	    int svalid, notbus, idvalid, luvalid, lun;
	    const char *scs = NULL;
	    
	    ++i;
	    addr = (buf[p+0]<<8) + buf[p+1];
	    if (element && element != addr)
		continue;
		
	    switch (etype)
	    {
	      case 1: /* Medium Transport Element */
		fprintf(fp, "\n    Transport #%u:\n", n);
		fprintf(fp, "      Element Adress:                 %u\n", addr);
		fprintf(fp, "      Abnormal State:                 %s\n", (buf[p+2] & 0x04) ? "Yes" : "No");
		fprintf(fp, "      Full:                           %s\n", (buf[p+2] & 0x01) ? "Yes" : "No");

		scs = scsi_sense_code2string(buf[p+4], buf[p+5]);
		fprintf(fp, "      Sense Code:                     0x%02x / 0x%02x (%s)\n",
		       buf[p+4], buf[p+5], scs ? scs : "Unknown");
			
		svalid = (buf[p+9]&0x80);
		if (svalid)
		{
		    fprintf(fp, "      Element Inverted:               %s\n", (buf[p+9]&0x40) ? "Yes" : "No");
		    fprintf(fp, "      Source Storage Element Address: %u\n", (buf[p+10]<<8) + buf[p+11]);
		}
		xp = p+12;
		if (pvoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Primary Volume Tag:             %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Primary Volume Tag:             %.32s\n", buf+xp);
		    xp += 36;
		}
		if (avoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Alternate Volume Tag:           %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Alternate Volume Tag:           %.32s\n", buf+xp);
		    xp += 36;
		}
		break;

	      case 2: /* Storage Element */
		fprintf(fp, "\n    Storage #%u:\n", n);
		fprintf(fp, "      Element Adress:                 %u\n", addr);
		fprintf(fp, "      Access State:                   %s\n", (buf[p+2] & 0x08) ? "Yes" : "No");
		fprintf(fp, "      Abnormal State:                 %s\n", (buf[p+2] & 0x04) ? "Yes" : "No");
		fprintf(fp, "      Full:                           %s\n", (buf[p+2] & 0x01) ? "Yes" : "No");
		scs = scsi_sense_code2string(buf[p+4], buf[p+5]);
		fprintf(fp, "      Sense Code:                     0x%02x / 0x%02x (%s)\n",
		       buf[p+4], buf[p+5], scs ? scs : "Unknown");
			
		svalid = (buf[p+9]&0x80);
		if (svalid)
		{
		    fprintf(fp, "      Element Inverted:               %s\n", (buf[p+9]&0x40) ? "Yes" : "No");
		    fprintf(fp, "      Source Storage Element Address: %u\n", (buf[p+10]<<8) + buf[p+11]);
		}
		xp = p+12;
		if (pvoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Primary Volume Tag:             %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Primary Volume Tag:             %.32s\n", buf+xp);
		    xp += 36;
		}
		if (avoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Alternate Volume Tag:           %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Alternate Volume Tag:           %.32s\n", buf+xp);
		    xp += 36;
		}
		break;

	      case 3: /* Import/Export Element */
		fprintf(fp, "\n    IOPort #%u:\n", n);
		fprintf(fp, "      Element Adress:                 %u\n", addr);
		fprintf(fp, "      Import Enabled:                 %s\n", (buf[p+2] & 0x20) ? "Yes" : "No");
		fprintf(fp, "      Export Enabled                  %s\n", (buf[p+2] & 0x10) ? "Yes" : "No");
		fprintf(fp, "      Access State:                   %s\n", (buf[p+2] & 0x08) ? "Yes" : "No");
		fprintf(fp, "      Abnormal State:                 %s\n", (buf[p+2] & 0x04) ? "Yes" : "No");
		fprintf(fp, "      Full:                           %s\n", (buf[p+2] & 0x01) ? "Yes" : "No");
		scs = scsi_sense_code2string(buf[p+4], buf[p+5]);
		fprintf(fp, "      Sense Code:                     0x%02x / 0x%02x (%s)\n",
		       buf[p+4], buf[p+5], scs ? scs : "Unknown");
			
		svalid = (buf[p+9]&0x80);
		if (svalid)
		{
		    fprintf(fp, "      Element Inverted:               %s\n", (buf[p+9]&0x40) ? "Yes" : "No");
		    fprintf(fp, "      Source Storage Element Address: %u\n", (buf[p+10]<<8) + buf[p+11]);
		}
		xp = p+12;
		if (pvoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Primary Volume Tag:             %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Primary Volume Tag:             %.32s\n", buf+xp);
		    xp += 36;
		}
		if (avoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Alternate Volume Tag:           %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Alternate Volume Tag:           %.32s\n", buf+xp);
		    xp += 36;
		}
		break;

	      case 4: /* Data Transfer Element */
		fprintf(fp, "\n    Drive #%u:\n", n);
		fprintf(fp, "      Element Adress:                 %u\n", addr);
		fprintf(fp, "      Access State:                   %s\n", (buf[p+2] & 0x08) ? "Yes" : "No");
		fprintf(fp, "      Abnormal State:                 %s\n", (buf[p+2] & 0x04) ? "Yes" : "No");
		fprintf(fp, "      Full:                           %s\n", (buf[p+2] & 0x01) ? "Yes" : "No");
		scs = scsi_sense_code2string(buf[p+4], buf[p+5]);
		fprintf(fp, "      Sense Code:                     0x%02x / 0x%02x (%s)\n",
		       buf[p+4], buf[p+5], scs ? scs : "Unknown");
			
		notbus = (buf[p+6]&0x80);
		idvalid = (buf[p+6]&0x20);
		luvalid = (buf[p+6]&0x10);
		lun = (buf[p+6]&0x07);

		fprintf(fp, "      Not on same SCSI Bus:           %s\n", notbus ? "Yes" : "No");
		if (idvalid)
		    fprintf(fp, "      SCSI Bus Address:               %u\n", buf[p+7]);
		if (luvalid)
		    fprintf(fp, "      SCSI Bus Logical Unit:          %u\n", lun);
			    
		svalid = (buf[p+9]&0x80);
			
		if (svalid)
		{
		    fprintf(fp, "      Element Inverted:               %s\n", (buf[p+9]&0x40) ? "Yes" : "No");
		    fprintf(fp, "      Source Storage Element Address: %u\n", (buf[p+10]<<8) + buf[p+11]);
		}
		xp = p+12;
		if (pvoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Primary Volume Tag:             %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Primary Volume Tag:             %.32s\n", buf+xp);
		    xp += 36;
		}
		if (avoltag)
		{
		    seq = (buf[xp+34]<<8)+buf[xp+35];
		    if (seq)
			fprintf(fp, "      Alternate Volume Tag:           %.32s [#%u]\n", buf+xp, seq);
		    else
			fprintf(fp, "      Alternate Volume Tag:           %.32s\n", buf+xp);
		    xp += 36;
		}
		break;
	      default:
		fprintf(fp, "\n    Unknown Object #%u:\n", n);
	    }
	}
    }
}

void
print_element_status(FILE *fp,
		     unsigned char *buf,
		     size_t buflen,
		     unsigned int element)
{
    int p = 0;
    int i, j, elemtypes;
    int n, addr, svalid, src;
    
    
    elemtypes = (buf[p+2]<<8) + buf[p+3];
    
    p += 8;
    for (i = 0; i < elemtypes;)
    {
	size_t elength, dlength;
	int etype, pvoltag, avoltag;
	
	etype = buf[p];
	pvoltag = (buf[p+1]&0x80);
	avoltag = (buf[p+1]&0x40);
	

	
	elength = (buf[p+2]<<8) + buf[p+3];
	dlength = (buf[p+5]<<16) + (buf[p+6]<<8) + buf[p+7];
	
	p += 8;
	n = element_start;
	
	for (j = 0; j < dlength; j += elength, p += elength, n++)
	{
	    int xp;
	    const char *scs = NULL;
	    
	    addr   = (buf[p+0]<<8) + buf[p+1];
	    svalid = (buf[p+9]&0x80);
	    src    = (buf[p+10]<<8)+buf[p+11];
	    
	    ++i;
	    if (element && element != addr)
		continue;

	    fprintf(fp, "%6u", addr);
	    
	    switch (etype)
	    {
	      case 1: /* Medium Transport Element */
		fprintf(fp, "\t%9s:%u\t", "transport", n);
		break;

	      case 2: /* Storage Element */
		fprintf(fp, "\t%9s:%u\t", "storage", n);
		break;

	      case 3: /* Import/Export Element */
		fprintf(fp, "\t%9s:%u\t", "ioport", n);
		break;

	      case 4: /* Data Transfer Element */
		fprintf(fp, "\t%9s:%u\t", "drive", n);
		break;

	      default:
		fprintf(fp, "\t%9s[%u]:%u\t", "unknown", etype, n);
	    }
	    
	    if ((buf[p+2] & 0x04) != 0)
	    {
		fprintf(fp, "fault    ");
		scs = scsi_sense_code2string(buf[p+4], buf[p+5]);
		fprintf(fp, "\tsense 0x%02x / 0x%02x (%s)\n",
			buf[p+4], buf[p+5], scs ? scs : "Unknown");
	    }
	    else
	    {
		printf("%s", (buf[p+2] & 0x01) ? "full" : "empty");
		if (svalid && src != addr)
		    fprintf(fp, "[%u] ", src);
		else
		    fprintf(fp, "    ");
		
		xp = p+12;
		if (pvoltag)
		{
		    fprintf(fp, "\t%.32s", buf+xp);
		    xp += 36;
		}
		if (avoltag)
		{
		    fprintf(fp, "\t%.32s", buf+xp);
		    xp += 36;
		}
		putc('\n', fp);
	    }
	}
    }
}



void
print_mode_page_parameter_header(FILE *fp,
				 unsigned char *buf)
{
    /* Mode Parameter Header (6) */
    fprintf(fp, "Mode Data Length:          %u\n", buf[0]);
    fprintf(fp, "Medium Type:               %u\n", buf[1]);
    fprintf(fp, "Device-Specific Parameter: %u\n", buf[2]);
    fprintf(fp, "Block Descriptor Length:   %u\n", buf[3]);
}


void
print_mode_page_element_address_assignment(FILE *fp,
					   unsigned char *buf,
					   size_t buflen)
{
    if (buflen != 20)
	return;

    /* Mode Page Element Address Assignment */
    fprintf(fp, "Parameters Savable:              %s\n", (buf[0]&0x80) ? "Yes" : "No");
    fprintf(fp, "Parameter Length:                %u\n", buf[1]);
    fprintf(fp, "Medium Transport Element Start:  %u\n", (buf[2]<<8)|buf[3]);
    fprintf(fp, "Medium Transport Element Count:  %u\n", (buf[4]<<8)|buf[5]);
    fprintf(fp, "Storage Element Start:           %u\n", (buf[6]<<8)|buf[7]);
    fprintf(fp, "Storage Element Count:           %u\n", (buf[8]<<8)|buf[9]);
    fprintf(fp, "Import/Export Element Start:     %u\n", (buf[10]<<8)|buf[11]);
    fprintf(fp, "Import/Export Element Count:     %u\n", (buf[12]<<8)|buf[13]);
    fprintf(fp, "Data Transfer Element Start:     %u\n", (buf[14]<<8)|buf[15]);
    fprintf(fp, "Data Transfer Element Count:     %u\n", (buf[16]<<8)|buf[17]);
}
