/*
** scsi.c - Generic SCSI command interface
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include "pjc.h"


const char *
scsi_status2string(int status)
{
    switch (status)
    {
      case SCSI_STATUS_GOOD:
	return "Good";
	
      case SCSI_STATUS_CHECK_CONDITION:
	return "Check Condition";
	
      case SCSI_STATUS_CONDITION_MET_GOOD:
	return "Condition Met/Good";
	
      case SCSI_STATUS_BUSY:
	return "Busy";
	
      case SCSI_STATUS_INTERMEDIATE_GOOD:
	return "Intermediate/Good";
	
      case SCSI_STATUS_INTERMEDIATE_CONDITOIN_MET_GOOD:
	return "Intermediate/Condition Met/Good";
	
      case SCSI_STATUS_RESERVATION_CONFLICT:
	return "Reservation Conflict";
	
      case SCSI_STATUS_COMMAND_TERMINATED:
	return "Command Terminated";
	
      case SCSI_STATUS_QUEUE_FULL:
	return "Queue Full";

      default:
	return NULL;
    }
}
	

const char *
scsi_devicetype2string(int dt)
{
    switch (dt)
    {
      case 0x00:
	return "Disk Drive";
      case 0x01:
	return "Tape Drive";
      case 0x02:
	return "Printer";
      case 0x03:
	return "Processor Device";
      case 0x04:
	return "WORM Drive";
      case 0x05:
	return "CD/DVD Drive";
      case 0x06:
	return "Scanner";
      case 0x07:
	return "Optical Disk";
      case 0x08:
	return "Media Changer";
      case 0x09:
	return "Communications Device";
      case 0x0A:
      case 0x0B:
	return "Printer Preprocessing Device";
      case 0x0C:
	return "Storage Array Controller";
      case 0x0D:
	return "Storage Enclosure Services";
      case 0x0E:
	return "Simplified Direct-Access Device";
      case 0x0F:
	return "Optical Card Reader/Writer Device";
      case 0x10:
	return "Bridge Controller Device";
      case 0x11:
	return "Object-based Storage Device";
      case 0x12:
	return "Automation/Drive Interface";
      case 0x1E:
	return "Well known logical unit";
      case 0x1F:
	return "Unknown or no device type";
      default:
	return "Reserved";
    }
}

const char *
scsi_ansiversion2string(int av)
{
    switch (av)
    {
      case 0:
	return "SCSI-1";
      case 1:
	return "CCS";
      case 2:
	return "SCSI-2";
      case 3:
	return "SCSI-3 (SPC)";
      case 4:
	return "SCSI-3 (SPC-2)";
      case 5:
	return "SCSI-3 (SPC-3)";
      default:
	return "Unknown";
    }
}


const char *
scsi_sense_error2string(int ec)
{
    switch (ec)
    {
      case 0x70:
	return "Current Error - DAL Block NOT From Sense Data";

      case 0x71:
	return "Deferred Error - DAL Block NOT From Sense Data";

      case 0xF0:
	return "Current Error - DAL Block IS From Sense Data";

      case 0xF1:
	return "Deferred Error - DAL Block IS From Sense Data";

      case 0x7F:
	return "Vendor Specific Error - DAL Block NOT From Sense Data";

      case 0xFF:
	return "Vendor Specific Error - DAL Block IS From Sense Data";

      default:
	return NULL;
    }
}


const char *
scsi_sense_key2string(int sk)
{
    switch (sk&0x0F)
    {
      case SCSI_SENSE_KEY_NO_SENSE:
	return "No Sense";
      case SCSI_SENSE_KEY_RECOVERED_ERROR:
	return "Recovered Error";
      case SCSI_SENSE_KEY_NOT_READY:
	return "Not Ready";
      case SCSI_SENSE_KEY_MEDIUM_ERROR:
	return "Medium Error";
      case SCSI_SENSE_KEY_HARDWARE_ERROR:
	return "Hardware Error";
      case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
	return "Illegal Request";
      case SCSI_SENSE_KEY_UNIT_ATTENTION:
	return "Unit Attention";
      case SCSI_SENSE_KEY_DATA_PROTECT:
	return "Data Protect";
      case SCSI_SENSE_KEY_BLANK_CHECK:
	return "Blank Check";
      case SCSI_SENSE_KEY_VENDOR_SPECIFIC:
	return "Vendor Specific";
      case SCSI_SENSE_KEY_COPY_ABORTED:
	return "Copy Aborted";
      case SCSI_SENSE_KEY_ABORTED_COMMAND:
	return "Aborted Command";
      case SCSI_SENSE_KEY_EQUAL:
	return "Equal";
      case SCSI_SENSE_KEY_VOLUME_OVERFLOW:
	return "Volume Overflow";
      case SCSI_SENSE_KEY_MISCOMPARE:
	return "Miscompare";
      default:
	return "Reserved/Unknown";
    }
}

const char *
scsi_sense_code2string(int sc, int asc)
{
    switch (((sc&0xFF)<<8)|(asc&0xFF))
    {
      case 0x0000:
	return "No additional sense information";
	
      case 0x0001:
	return "Filemark detected";
	
      case 0x0002:
	return "End of partition/medium deteced";
	
      case 0x0003:
	return "Setmark detected";
	
      case 0x0004:
	return "Beginning of partition/medium detected";

      case 0x0005:
	return "End of data detected";

      case 0x0006:
	return "I/O process terminated";

      case 0x0011:
	return "Audio play operation in progress";

      case 0x0012:
	return "Audio play operation paused";

      case 0x0013:
	return "Audio play operation successfully completed";

      case 0x0014:
	return "Audio play operation stopped due to error";

      case 0x0015:
	return "No current audio status to return";

      case 0x0100:
	return "No index / sector signal";

      case 0x0200:
	return "No seek complete";

      case 0x0300:
	return "Peripheral device write fault";

      case 0x0301:
	return "No write current";

      case 0x0302:
	return "Excessive write errors";

      case 0x0400:
	return "Logical unit not ready, cause not reportable";

      case 0x0401:
	return "Logical unit is in progress of becoming ready";

      case 0x0402:
	return "Logical unit is not ready, initalizing command required";

      case 0x0403:
	return "Logical unit is not ready, manual intervention required";

      case 0x0404:
	return "Logical unit is not ready, format in progress";

      case 0x0500:
	return "Logical unit does not respond to selection";

      case 0x0600:
	return "Reference position found";

      case 0x0700:
	return "Multiple peripheral devices selected";

      case 0x0800:
	return "Logical unit communication failure";

      case 0x0801:
	return "Logical unit communication timeout";

      case 0x0802:
	return "Logical unit communication parity error";

      case 0x0900:
	return "Track following error";

      case 0x0901:
	return "Tracking servo failure";

      case 0x0902:
	return "Focus servo failure";

      case 0x0903:
	return "Spindle servo failure";

      case 0x0A00:
	return "Error log overflow";

      case 0x0C00:
	return "Write error";
	
      case 0x0C01:
	return "Write error recovered with auto reallocation";
	
      case 0x0C02:
	return "Write error auto reallocation failed";
	
      case 0x1000:
	return "ID CRC or ECC error";

      case 0x1100:
	return "Unrecovered read error";

      case 0x1101:
	return "Read retries exhausted";

      case 0x1102:
	return "Error too long to correct";

      case 0x1103:
	return "Multiple read errors";

      case 0x1104:
	return "Unrecovered read error - auto reallocate failed";

      case 0x1105:
	return "L-EC uncorrectable error";

      case 0x1106:
	return "CIRC unrecovered error";

      case 0x1107:
	return "Data resynchronization error";

      case 0x1108:
	return "Incomplete block read";

      case 0x1109:
	return "No gap found";

      case 0x110A:
	return "Miscorrected error";

      case 0x110B:
	return "Unrecovered read error - recommend reassignment";

      case 0x110C:
	return "Unrecovered read error - recommend re-write the data";

      case 0x1200:
	return "Address mark not found for ID field";

      case 0x1300:
	return "Address mark not found for data field";

      case 0x1400:
	return "Recorded entity not found";

      case 0x1401:
	return "Record not found";

      case 0x1402:
	return "Filemark or setmark not found";

      case 0x1403:
	return "End-of-data not found";

      case 0x1404:
	return "Block sequence error";

      case 0x1500:
	return "Random positioning error";

      case 0x1501:
	return "Mechanical positioning error";

      case 0x1502:
	return "Positioning error detected by read of medium";

      case 0x1600:
	return "Data synchronization mark error";

      case 0x1700:
	return "Recovered data with no error correction applied";

      case 0x1701:
	return "Recovered data with retries";

      case 0x1702:
	return "Recovered data with positive head offset";

      case 0x1703:
	return "Recovered data with negative head offset";

      case 0x1704:
	return "Recovered data with retries and/or CIRC applied";

      case 0x1705:
	return "Recovered data using previous sector ID";

      case 0x1706:
	return "Recovered data without ECC - data auto-reallocated";

      case 0x1707:
	return "Recovered data without ECC - recommend reassignment";

      case 0x1708:
	return "Recovered data without ECC - recommend rewrite";

      case 0x1800:
	return "Recovered data with error correction applied";

      case 0x1801:
	return "Recovered data with error correction & retries applied";

      case 0x1802:
	return "Recovered data - data auto-reallocated";

      case 0x1803:
	return "Recovered data with CIRC";

      case 0x1804:
	return "Recovered data with LEC";

      case 0x1805:
	return "Recovered data - recommend reassignment";

      case 0x1806:
	return "Recovered data - recommend rewrite";

      case 0x1900:
	return "Defect list error";

      case 0x1901:
	return "Defect list not available";

      case 0x1902:
	return "Defect list error in primary list";

      case 0x1903:
	return "Defect list error in grown list";

      case 0x1A00:
	return "Parameter list length error";

      case 0x1B00:
	return "Synchronous data transfer error";

      case 0x1C00:
	return "Defect list not found";

      case 0x1C01:
	return "Primary defect list not found";

      case 0x1C02:
	return "Grown defect list not found";

      case 0x1D00:
	return "Miscompare during verify operation";

      case 0x1E00:
	return "Recovered ID with ECC";

      case 0x2000:
	return "Invalid command operation code";

      case 0x2100:
	return "Logical block address out of range";

      case 0x2101:
	return "Invalid element address";

      case 0x2200:
	return "Illegal function (should use 20 00, 24 00, or 26 00)";

      case 0x2400:
	return "Invalid field in CDB";

      case 0x2500:
	return "Logical unit not supported";

      case 0x2600:
	return "Invalid field in parameter list";

      case 0x2601:
	return "Parameter not supported";

      case 0x2602:
	return "Parameter value invalid";

      case 0x2603:
	return "Threshold parameters not supported";

      case 0x2700:
	return "Write protected";

      case 0x2800:
	return "Not ready to ready transition (medium may have changed)";

      case 0x2801:
	return "Import or export element accessed";

      case 0x2900:
	return "Power on, Reset, or Bus device reset occurred";

      case 0x2A00:
	return "Parameters changed";

      case 0x2A01:
	return "Mode parameters changed";

      case 0x2A02:
	return "Log parameters changed";

      case 0x2B00:
	return "Copy cannot execute since host cannot disconnect";

      case 0x2C00:
	return "Command sequence error";

      case 0x2C01:
	return "Too many windows specified";

      case 0x2C02:
	return "Invalid combination of windows specified";

      case 0x2D00:
	return "Overwrite error on update in place";

      case 0x2F00:
	return "Commands cleared by another initiator";

      case 0x3000:
	return "Incompatible medium installed";

      case 0x3001:
	return "Cannot read medium - unknown format";

      case 0x3002:
	return "Cannot read medium - incompatible format";

      case 0x3003:
	return "Cleaning cartridge installed";

      case 0x3100:
	return "Medium format corrupted";

      case 0x3101:
	return "Format command failed";

      case 0x3200:
	return "No defect spare location available";

      case 0x3201:
	return "Defect list update failure";

      case 0x3300:
	return "Tape Length error";

      case 0x3600:
	return "Ribbon, ink or toner failure";

      case 0x3700:
	return "Rounded parameter";

      case 0x3900:
	return "Saving parameters not supported";

      case 0x3A00:
	return "Medium not present";

      case 0x3B00:
	return "Sequential Positioning error";

      case 0x3B01:
	return "Tape position error at beginning-of-medium";

      case 0x3B02:
	return "Tape position error at end-of-medium";

      case 0x3B03:
	return "Tape or electronic vertical forms unit not ready";

      case 0x3B04:
	return "Slew failure";

      case 0x3B05:
	return "Paper jam";

      case 0x3B06:
	return "Failed to sense top-of-form";

      case 0x3B07:
	return "Failed to sense bottom-of-form";

      case 0x3B08:
	return "Reposition error";

      case 0x3B09:
	return "Read past end-of-medium";

      case 0x3B0A:
	return "Read past beginning-of-medium";

      case 0x3B0B:
	return "Position past end-of-medium";

      case 0x3B0C:
	return "Position past beginning-of-medium";

      case 0x3B0D:
	return "Medium destination element full";

      case 0x3B0E:
	return "Medium source element empty";

      case 0x3D00:
	return "Medium destination element full";

      case 0x3E00:
	return "Logical unit has not self-configured yet";

      case 0x3F00:
	return "Target operating conditions have changed";

      case 0x3F01:
	return "Microcode has been changed";

      case 0x3F02:
	return "Changed operating definition";

      case 0x3F03:
	return "Inquiry data has changed";

      case 0x4000:
	return "RAM failure";

#if 0
      case 0x40NN:
	return "Diagnostic failure on component NN (80h-FFh)";
#endif
	
      case 0x4100:
	return "Data path failure";

      case 0x4200:
	return "Power-on or self-test failure";

      case 0x4300:
	return "Message error";

      case 0x4400:
	return "Internal target failure";

      case 0x4500:
	return "Select or reselect failure";

      case 0x4600:
	return "Unsuccessful soft reset";

      case 0x4700:
	return "SCSI parity error";

      case 0x4800:
	return "Initiator detected error message received";

      case 0x4900:
	return "Invalid message error";

      case 0x4A00:
	return "Command phase error";

      case 0x4B00:
	return "Data phase error";

      case 0x4C00:
	return "Logical unit failed self-configuration";

      case 0x4E00:
	return "Overlapped commands attempted";

      case 0x5000:
	return "Write append error";

      case 0x5001:
	return "Write append position error";

      case 0x5002:
	return "Position error related to timing";

      case 0x5100:
	return "Erase failure";

      case 0x5200:
	return "Cartridge fault";

      case 0x5300:
	return "Media load or eject failed";

      case 0x5301:
	return "Unload tape failure";

      case 0x5302:
	return "Medium removal prevented";

      case 0x5400:
	return "SCSI to host system interface failure";

      case 0x5500:
	return "System resource failure";

      case 0x5700:
	return "Unable to recover table-of-contents";

      case 0x5800:
	return "Generation does not exist";

      case 0x5900:
	return "Updated block read";

      case 0x5A00:
	return "Operator request or state change input (unspecified)";

      case 0x5A01:
	return "Operator medium removal request";

      case 0x5A02:
	return "Operator selected write protect";

      case 0x5A03:
	return "Operator selected write permit";

      case 0x5B00:
	return "Log exception";

      case 0x5B01:
	return "Threshold condition met";

      case 0x5B02:
	return "Log counter at maximum";

      case 0x5B03:
	return "Log list codes exhausted";

      case 0x5C00:
	return "RPL status change";

      case 0x5C01:
	return "Spindles synchronize";

      case 0x5C02:
	return "Spindles not synchronized";

      case 0x6000:
	return "Lamp Failure";

      case 0x6100:
	return "Video acquisition error";

      case 0x6101:
	return "Unable to acquire video";

      case 0x6102:
	return "Out of focus";

      case 0x6200:
	return "Scan head positioning error";

      case 0x6300:
	return "End of user area encountered on this track";

      case 0x6400:
	return "Illegal mode for this track";
	
      default:
	return NULL;
    }
}



const char *
scsi_element_type2string(int et)
{
    switch (et)
    {
      case 1:
	return "Medium Transport Element";

      case 2:
	return "Storage Element";

      case 3:
	return "Import/Export Element";

      case 4:
	return "Data Transfer Element";

      default:
	return NULL;
    }
}



SCSI *
scsi_open(const char *path)
{
    SCSI *sp;
    int rc, attempt;
    

    sp = malloc(sizeof(*sp));
    if (!sp)
	return NULL;

    memset(sp, 0, sizeof(*sp));

    attempt = 0;
  Retry:
    ++attempt;
    
    sp->fd = scsi_real_open(path);
    if (sp->fd < 0)
    {
	if (waitretry && errno == EBUSY)
	{
	    if (debug)
		fprintf(stderr, "scsi_open: device busy (errno=%d)\n", errno);
	    
	    if (verbose && attempt == 1)
		printf("[Waiting for device to become ready]\n");
	    
	    if (attempt >= 15)
		sleep(30);
	    else if (attempt >= 10)
		sleep(10);
	    else
		if (attempt > 1)
		    sleep(1);
	    
	    goto Retry;
	}
	
	goto Fail;
    }
    
    sp->path = strdup(path);
    sp->timeout = 0;

    rc = scsi_inquiry(sp, &sp->inquiry);
    if (rc != CMD_OK)
	goto Fail;

    sp->devtype = SCSI_INQUIRY_DEVTYPE(&sp->inquiry);
    sp->version = SCSI_INQUIRY_VERSION(&sp->inquiry);

    if (sp->devtype == SCSI_INQUIRY_DEVTYPE_CHANGER)
	sp->attached = 0;
    else if (sp->version >= 3 && SCSI_INQUIRY_MCHNGR(&sp->inquiry))
    {
	if (debug)
	    fprintf(stderr, "scsi_open(\"%s\"): ATTACHED CHANGER\n", path);
	
	sp->attached = 1;
    }

    return sp;

  Fail:
    if (sp->path)
	free((void *) sp->path);
    
    free(sp);
    return NULL;
}


int
scsi_close(SCSI *sp)
{
    int rc;


    rc = scsi_real_close(sp->fd);
    if (rc < 0)
	return rc;
    
    free((void *) sp->path);
    free(sp);
    
    return 0;
}

int
scsi_timeout(SCSI *sp,
	     int timeout)
{
    int old_timeout;

    old_timeout = sp->timeout;
    sp->timeout = timeout;

    return old_timeout;
}


int
scsi_test_unit_ready(SCSI *sp)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_TEST_UNIT_READY;

    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			10); /* 10 seconds - this command should be quick */
}


int
scsi_rezero_unit(SCSI *sp)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_REZERO_UNIT;

    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			600); /* 10 minutes - this command may take a long time */
}


int
scsi_initialize_element_status(SCSI *sp)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_INITIALIZE_ELEMENT_STATUS;

    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			600); /* 10 minutes - this command may take a long time */
}


int
scsi_request_sense(SCSI *sp,
		   SCSI_SENSE *ssp)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_REQUEST_SENSE;
    cdb[4] = 0xFF;

    return scsi_command(sp, SCSI_READ,
			&cdb, sizeof(cdb),
			ssp->buf, sizeof(ssp->buf), &ssp->len,
			10); /* 10 seconds - this command should be quick */
}


int
scsi_inquiry(SCSI *sp,
	     SCSI_INQUIRY *sip)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_INQUIRY;
    cdb[4] = 0xFF;

    return scsi_command(sp, SCSI_READ,
			&cdb, sizeof(cdb),
			sip->buf, sizeof(sip->buf), &sip->len,
			10); /* 10 seconds - this command should be quick */
}



int
scsi_reserve(SCSI *sp,
	     unsigned int *xid,    /* 3rdPty Device ID */
	     unsigned int *ersvid, /* Reservation ID */
	     void *evbuf,
	     size_t evsize)
	     
{
    CDB6 cdb;
    int rc;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_RESERVE;
    if (xid)
	cdb[1] |= (0x10 | ((*xid)&0x07)<<1);

    
    if (ersvid)
	cdb[2] = *ersvid;

    if (evbuf)
    {
	cdb[1] |= 0x01;
	rc = scsi_command(sp, SCSI_WRITE,
			  &cdb, sizeof(cdb),
			  evbuf, evsize, NULL,
			  10); /* 10 seconds - this should be quick */
    }
    else
	rc = scsi_command(sp, SCSI_NONE,
			  &cdb, sizeof(cdb),
			  NULL, 0, NULL,
			  10); /* 10 seconds - this should be quick */
    
    return rc;
}


int
scsi_release(SCSI *sp,
	     unsigned int *xid,     /* 3rdPty Device ID */
	     unsigned int *ersvid)  /* Reservation ID */
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_RELEASE;
    if (xid)
	cdb[1] |= (0x10 | ((*xid)&0x07)<<1);
    
    if (ersvid)
    {
	cdb[1] |= 0x01;
	cdb[2] = *ersvid;
    }
    
    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			10); /* 10 seconds - this should be quick */
}


int
scsi_move_medium(SCSI *sp,
		 unsigned int transport,
		 unsigned int source,
		 unsigned int destination,
		 int invert)
{
    CDB12 cdb;

    
    memset(&cdb, 0, sizeof(cdb));

    if (sp->attached)
	cdb[0] = SCSI_COMMAND_MOVE_MEDIUM_ATTACHED;
    else
	cdb[0] = SCSI_COMMAND_MOVE_MEDIUM;
    
    cdb[2] = (transport>>8)&0xFF;
    cdb[3] = transport&0xFF;

    cdb[4] = (source>>8)&0xFF;
    cdb[5] = source&0xFF;

    cdb[6] = (destination>>8)&0xFF;
    cdb[7] = destination&0xFF;

    cdb[10] |= (invert & 0x01);
    
    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			60); /* 1 minute - this should be medium quick */
}




int
scsi_read_element_status(SCSI *sp,
			 unsigned int ecode,
			 unsigned int start,
			 unsigned int elems,
			 void *buf,
			 size_t buflen,
			 size_t *bufres)
{
    CDB12 cdb;

    
    memset(&cdb, 0, sizeof(cdb));

    if (sp->attached)
	cdb[0] = SCSI_COMMAND_READ_ELEMENT_STATUS_ATTACHED;
    else
	cdb[0] = SCSI_COMMAND_READ_ELEMENT_STATUS;
    
    cdb[1] |= 0x10; /* VolTag */
    cdb[1] |= (ecode & 0x0F);
    
    cdb[2] = ((start >> 8) & 0xFF);
    cdb[3] = (start & 0xFF);
    
    cdb[4] = ((elems >> 8) & 0xFF);
    cdb[5] = (elems & 0xFF);
    
    cdb[7] = ((buflen >> 16) & 0xFF);
    cdb[8] = ((buflen >> 8) & 0xFF);
    cdb[9] = (buflen & 0xFF);
    
    return scsi_command(sp, SCSI_READ,
			&cdb, sizeof(cdb),
			buf, buflen, bufres,
			10); /* 10 seconds - this command should be quick */
}


int
scsi_mode_sense_6(SCSI *sp,
		  unsigned int dbd,
		  unsigned int pc,
		  unsigned int page_code,
		  void *buf,
		  size_t buflen,
		  size_t *bufres)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_MODE_SENSE_6;
    cdb[1] |= (dbd ? 0x08 : 0);
    cdb[2] |= ((pc&0x03)<<6)|(page_code & 0x3F);
    cdb[4] = buflen;
    
    return scsi_command(sp, SCSI_READ,
			&cdb, sizeof(cdb),
			buf, buflen, bufres,
			10); /* 10 seconds - this command should be quick */
}


int
scsi_position_to_element(SCSI *sp,
			 unsigned int transport,
			 unsigned int destination,
			 int invert)
{
    CDB10 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_POSITION_TO_ELEMENT;
    
    cdb[2] = (transport>>8)&0xFF;
    cdb[3] = transport&0xFF;

    cdb[4] = (destination>>8)&0xFF;
    cdb[5] = destination&0xFF;

    cdb[8] |= (invert ? 0x01 : 0);
    
    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			60); /* 1 minute - this should be medium quick */
}


int
scsi_exchange_medium(SCSI *sp,
		     unsigned int transport,
		     unsigned int source,
		     unsigned int destination1,
		     unsigned int destination2,
		     int invert)
{
    CDB12 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_EXCHANGE_MEDIUM;
    
    cdb[2] = (transport>>8)&0xFF;
    cdb[3] = transport&0xFF;

    cdb[4] = (source>>8)&0xFF;
    cdb[5] = source&0xFF;

    cdb[6] = (destination1>>8)&0xFF;
    cdb[7] = destination1&0xFF;

    cdb[8] = (destination2>>8)&0xFF;
    cdb[9] = destination2&0xFF;

    cdb[10] |= (invert&0x03);
    
    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			60); /* 1 minute - this should be medium quick */
}


int
scsi_prevent_allow_removal(SCSI *sp,
			   int prevent)
{
    CDB6 cdb;

    
    memset(&cdb, 0, sizeof(cdb));
    
    cdb[0] = SCSI_COMMAND_PREVENT_ALLOW_MEDIUM_REMOVAL;
    cdb[4] = prevent;

    return scsi_command(sp, SCSI_NONE,
			&cdb, sizeof(cdb),
			NULL, 0, NULL,
			10); /* 10 seconds - this should be quick */
}



/*
** Perform a generic SCSI operation.
**
** Return values:
**   0 : Operation Successful
**  <0 : ioctl() syscall failed, value is -errno
**  >0 : SCSI Operation failed, value is SCSI Status Byte
*/
int
scsi_command(SCSI *sp,            /* File descriptor for SCSI device */
	     int dir,             /* I/O direction, SCSI_NONE, SCSI_READ, SCSI_WRITE */
	     void *cdb,           /* SCSI Command Descriptor Block */
	     size_t cdblen,       /* Length of CDB */
	     void *iobuf,         /* I/O Data Buffer */
	     size_t iobuflen,     /* Read: Max size of data to transfer, Write: Data to transfer */
	     size_t *iobufres,    /* Read: Size of data returned, Write: Data written */
	     int timeout)         /* Transaction timeout */
{
    int rc, attempt;
    
    
    if (debug)
	fprintf(stderr, "scsi_command(fd=%d, dir=%d, cdblen=%u, iolen=%u, timeout=%d)\n",
		sp->fd, dir, (unsigned int) cdblen, (unsigned int) iobuflen, timeout);
    
    if (pktdump)
	print_buffer(stderr, "Command Descriptor Block", cdb,
		     pktdump < cdblen ? pktdump : cdblen);
    
    if (pktdump && dir == SCSI_WRITE)
	print_buffer(stderr, "Write Data Block", iobuf,
		     pktdump < iobuflen ? pktdump : iobuflen);

    attempt = 0;
  Retry:
    ++attempt;
    
    rc = scsi_real_command(sp->fd, dir, cdb, cdblen,
			   iobuf, iobuflen, iobufres,
			   sp->sense.buf, sizeof(sp->sense.buf), &sp->sense.len,
			   sp->timeout ? sp->timeout : timeout);

    
    /* Retry temporary or 'Not Ready' errors? */
    if (waitretry &&
	(rc == SCSI_STATUS_CHECK_CONDITION || rc == SCSI_STATUS_COMMAND_TERMINATED) &&
 	sp->sense.len >= 14 && SCSI_SENSE_ERROR(&sp->sense) == 0x70 &&
	(SCSI_SENSE_KEY(&sp->sense) == SCSI_SENSE_KEY_NOT_READY ||
 	 (SCSI_SENSE_KEY(&sp->sense) == SCSI_SENSE_KEY_UNIT_ATTENTION &&
	  SCSI_SENSE_ASC(&sp->sense) == 0x29 && SCSI_SENSE_ASCQ(&sp->sense) == 0x00)) &&
	(waitretry < 0 || attempt < waitretry))
    {
	if (debug)
	    fprintf(stderr, "scsi_command: not ready (key 0x%02x code 0x%02x/0x%02x) [#%d]\n",
		    SCSI_SENSE_KEY(&sp->sense), 
		    SCSI_SENSE_ASC(&sp->sense),
		    SCSI_SENSE_ASCQ(&sp->sense),
		    attempt);

	if (verbose && attempt == 1)
	    printf("[Waiting for device to become ready]\n");
	
 	/* "Not Ready" or "Unit Attention" */
	if (attempt >= 15)
	    sleep(30);
	else if (attempt >= 10)
	    sleep(10);
	else
	    if (attempt > 1)
		sleep(1);
	
	goto Retry;
    }
		  
    if (debug)
	fprintf(stderr, "scsi_command: retval=%d (errno=%d)\n", rc, errno);
    
    if (pktdump && (dir == SCSI_READ && iobufres))
	print_buffer(stderr, "Read Data Block", iobuf, pktdump < *iobufres ? pktdump : *iobufres);
  
    if (pktdump && sp->sense.len > 0 && rc > 0)
	print_buffer(stderr, "Sense Data Block", sp->sense.buf,
		     pktdump < sp->sense.len ? pktdump : sp->sense.len);
    
    return rc;
}
