/*
** scsi.h - Generic SCSI command interface
**
** Copyright (c) 2005 Peter Eriksson <pen@lysator.liu.se>
*/

#ifndef PJC_SCSI_H
#define PJC_SCSI_H 1

#include <sys/types.h>

#define SCSI_NONE  0
#define SCSI_READ  1
#define SCSI_WRITE 2

typedef unsigned char CDB6[6];
typedef unsigned char CDB10[10];
typedef unsigned char CDB12[12];
typedef unsigned char CDB16[16];

typedef struct
{
    size_t len;
    unsigned char buf[255];
} SCSI_SENSE, SCSI_INQUIRY;

/*
** SCSI Medium Changer Commands
*/

#define SCSI_COMMAND_TEST_UNIT_READY                 0x00
#define SCSI_COMMAND_REZERO_UNIT                     0x01
#define SCSI_COMMAND_REQUEST_SENSE	             0x03
#define SCSI_COMMAND_INITIALIZE_ELEMENT_STATUS       0x07
#define SCSI_COMMAND_INQUIRY			     0x12

#define SCSI_COMMAND_CHANGE_DEFINITION               0x40
#define SCSI_COMMAND_EXCHANGE_MEDIUM                 0xA6
#define SCSI_COMMAND_LOG_SELECT                      0x4C
#define SCSI_COMMAND_LOG_SENSE                       0x4D
#define SCSI_COMMAND_MODE_SELECT_6                   0x15
#define SCSI_COMMAND_MODE_SELECT_10                  0x55
#define SCSI_COMMAND_MODE_SENSE_6                    0x1A
#define SCSI_COMMAND_MODE_SENSE_10                   0x5A
#define SCSI_COMMAND_MOVE_MEDIUM                     0xA5
#define SCSI_COMMAND_MOVE_MEDIUM_ATTACHED            0xA7
#define SCSI_COMMAND_POSITION_TO_ELEMENT             0x2B
#define SCSI_COMMAND_PREVENT_ALLOW_MEDIUM_REMOVAL    0x1E
#define SCSI_COMMAND_READ_BUFFER                     0x3C
#define SCSI_COMMAND_READ_ELEMENT_STATUS             0xB8
#define SCSI_COMMAND_READ_ELEMENT_STATUS_ATTACHED    0xB4
#define SCSI_COMMAND_RECEIVE_DIAGNOSTIC_RESULTS      0x1C
#define SCSI_COMMAND_RELEASE                         0x17
#define SCSI_COMMAND_REQUEST_VOLUME_ELEMENT_ADDRESS  0xB5
#define SCSI_COMMAND_RESERVE                         0x16
#define SCSI_COMMAND_SEND_DIAGNOSTIC                 0x1D
#define SCSI_COMMAND_SEND_VOLUME_TAG                 0xB6
#define SCSI_COMMAND_WRITE_BUFFER                    0x3B


#define SCSI_STATUS_GOOD                             0x00
#define SCSI_STATUS_CHECK_CONDITION                  0x02
#define SCSI_STATUS_CONDITION_MET_GOOD               0x04
#define SCSI_STATUS_BUSY                             0x08
#define SCSI_STATUS_INTERMEDIATE_GOOD                0x10
#define SCSI_STATUS_INTERMEDIATE_CONDITOIN_MET_GOOD  0x14
#define SCSI_STATUS_RESERVATION_CONFLICT             0x18
#define SCSI_STATUS_COMMAND_TERMINATED               0x22
#define SCSI_STATUS_QUEUE_FULL                       0x28

#define SCSI_MODE_PAGE_ELEMENT_ADDRESS_ASSIGNMENT    0x1D
#define SCSI_MODE_PAGE_TRANSPORT_GEOMETRY_PARAMETERS 0x1E
#define SCSI_MODE_PAGE_DEVICE_CAPABILITIES           0x1F

#define SCSI_SENSE_KEY_NO_SENSE                      0x00
#define SCSI_SENSE_KEY_RECOVERED_ERROR               0x01
#define SCSI_SENSE_KEY_NOT_READY                     0x02
#define SCSI_SENSE_KEY_MEDIUM_ERROR                  0x03
#define SCSI_SENSE_KEY_HARDWARE_ERROR                0x04
#define SCSI_SENSE_KEY_ILLEGAL_REQUEST               0x05
#define SCSI_SENSE_KEY_UNIT_ATTENTION                0x06
#define SCSI_SENSE_KEY_DATA_PROTECT                  0x07
#define SCSI_SENSE_KEY_BLANK_CHECK                   0x08
#define SCSI_SENSE_KEY_VENDOR_SPECIFIC               0x09
#define SCSI_SENSE_KEY_COPY_ABORTED                  0x0A
#define SCSI_SENSE_KEY_ABORTED_COMMAND               0x0B
#define SCSI_SENSE_KEY_EQUAL                         0x0C
#define SCSI_SENSE_KEY_VOLUME_OVERFLOW               0x0D
#define SCSI_SENSE_KEY_MISCOMPARE                    0x0E
#define SCSI_SENSE_KEY_RESERVED                      0x0F

#define SCSI_SENSE_ERROR(p)		((p)->buf[0]&0x7F)
#define SCSI_SENSE_KEY(p)		((p)->buf[2]&0x0F)
#define SCSI_SENSE_ASC(p)		((p)->buf[12])
#define SCSI_SENSE_ASCQ(p)		((p)->buf[13])
#define SCSI_SENSE_FRU(p)		((p)->buf[14])


#define SCSI_INQUIRY_VERSION(p)		( ((p)->buf[2]&0x07) < 3 ? ((p)->buf[2]&0x07) : ((p)->buf[2]) )
#define SCSI_INQUIRY_MCHNGR(p)		((p)->buf[6]&0x80)
#define SCSI_INQUIRY_DEVTYPE(p)		((p)->buf[0]&0x1F)

#define SCSI_INQUIRY_DEVTYPE_DISK	0x00
#define SCSI_INQUIRY_DEVTYPE_TAPE	0x01
#define SCSI_INQUIRY_DEVTYPE_CHANGER	0x08


typedef struct
{
    int fd;
    const char *path;
    int timeout;        /* Override default timeout values */
    
    int devtype;
    int version;
    int attached;
    
    SCSI_INQUIRY inquiry;
    SCSI_SENSE   sense;
} SCSI;


extern SCSI *
scsi_open(const char *dev);

extern int
scsi_close(SCSI *sp);

extern int
scsi_timeout(SCSI *sp,
	     int timeout);

const char *
scsi_status2string(int status);

extern const char *
scsi_devicetype2string(int dt);

extern const char *
scsi_ansiversion2string(int av);

extern const char *
scsi_sense_error2string(int ec);

extern const char *
scsi_sense_key2string(int sk);

extern const char *
scsi_sense_code2string(int sc, int asc);

extern const char *
scsi_element_type2string(int et);



/*
** Perform a generic SCSI operation.
**
** Return values:
**   0 : Operation Successful
**  <0 : ioctl() syscall failed, value is -errno
**  >0 : SCSI Operation failed, value is SCSI Status Byte
*/
extern int
scsi_command(SCSI *sp,            /* Descriptor for SCSI device */
	     int dir,             /* I/O direction, SCSI_NONE, SCSI_READ, SCSI_WRITE */
	     void *cdb,           /* SCSI Command Descriptor Block */
	     size_t cdblen,       /* Length of CDB */
	     void *iobuf,         /* I/O Data Buffer */
	     size_t iobuflen,     /* Read: Max size of data to transfer, Write: Data to transfer */
	     size_t *iobufres,    /* Read: Size of data returned, Write: Data written */
	     int timeout);         /* Transaction timeout in seconds */



extern int
scsi_test_unit_ready(SCSI *sp);

extern int
scsi_rezero_unit(SCSI *sp);

extern int
scsi_initialize_element_status(SCSI *sp);

extern int
scsi_request_sense(SCSI *sp,
		   SCSI_SENSE *ssp);

extern int
scsi_inquiry(SCSI *sp,
	     SCSI_INQUIRY *sip);


extern int
scsi_reserve(SCSI *sp,
	     unsigned int *xid,    /* 3rdPty Device ID */
	     unsigned int *ersvid, /* Reservation ID */
	     void *evbuf,
	     size_t evsize);

extern int
scsi_release(SCSI *sp,
	     unsigned int *xid,      /* 3rdPty Device ID */
	     unsigned int *ersvid);  /* Reservation ID */

extern int
scsi_move_medium(SCSI *sp,
		 unsigned int transport,
		 unsigned int source,
		 unsigned int destination,
		 int invert);

extern int
scsi_read_element_status(SCSI *sp,
			 unsigned int ecode,
			 unsigned int start,
			 unsigned int elems,
			 void *buf,
			 size_t bufsize,
			 size_t *buflen);

extern int
scsi_mode_sense_6(SCSI *sp,
		  unsigned int dbd,
		  unsigned int pc,
		  unsigned int page_code,
		  void *buf,
		  size_t buflen,
		  size_t *bufres);

extern int
scsi_position_to_element(SCSI *sp,
			 unsigned int transport,
			 unsigned int destination,
			 int invert);

extern int
scsi_exchange_medium(SCSI *sp,
		     unsigned int transport,
		     unsigned int source,
		     unsigned int destination1,
		     unsigned int destination2,
		     int invert);


extern int
scsi_prevent_allow_removal(SCSI *sp,
			   int prevent);





/* Low-level system-specific handlers */


extern int
scsi_real_open(const char *dev);

extern int
scsi_real_close(int fd);

extern int
scsi_real_command(int fd,              /* File descriptor for SCSI device */
		  int dir,             /* I/O direction, SCSI_NONE, SCSI_READ, SCSI_WRITE */
		  void *cdb,           /* SCSI Command Descriptor Block */
		  size_t cdblen,       /* Length of CDB */
		  void *iobuf,         /* I/O Data Buffer */
		  size_t iobuflen,     /* Read: Max size of data to transfer, Write: Data to transfer */
		  size_t *iobufres,    /* Read: Size of data returned, Write: Data written */
		  void *rqbuf,         /* Request Sense Buffer */
		  size_t rqbuflen,     /* Read: Max size of sense data to transfer */
		  size_t *rqbufres,    /* Read: Size of sense data returned */
		  int timeout);         /* Transaction timeout in seconds */


#endif
